#!/usr/bin/python
# Copyright (C) 2001,  Christof Meerwald <cmeerw@web.de>
# http://xpcom2corba.sunsite.dk/

# XPCOM-CORBA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

import md5, os, string, sys

from omniidl import idlast, idlvisitor, idlutil, idltype, output
from xpidl_tmpl import *
import xpcorba_type
convert_scoped2flat = xpcorba_type.convert_scoped2flat


REPOID_ISUPPORTS = 'DCE:00000000-0000-0000-c000-000000000046:1'


class XPIDL_Visitor(idlvisitor.AstVisitor):
    def __init__(self):
        self._idl = None
        self._included = ['xpcom.idl']

    def computeHexDigest(self, inputString):
        try:
           return md5.md5(inputString).hexdigest()
        except AttributeError:
           digest = '';
           for char in md5.md5(inputString).digest():
              digest = digest + ('%02x' % ord(char))
           return digest

    def visitAST(self, node):
        directory, basename = os.path.split(node.file())
        if string.lower(basename[-4:]) == '.idl':
            basename = basename[:-4]

        self._idl = output.Stream(open('I' + basename + '.idl', 'w'), 2)
        gbasename = '';
        for i in basename:
            if (i >= 'A' and i <= 'Z') or (i >= 'a' and i <= 'z'):
                gbasename = gbasename + i
        self._idl.out(TMPL_idl_file_header,
                      guardname='XPIDL__' + string.upper(gbasename) + '__INCLUDED')

        for n in node.declarations():
            if n.mainFile():
                n.accept(self)
            else:
                filename = n.file()
                pos = string.rfind(filename, '/')
                if pos != -1:
                    filename = filename[pos + 1:]

                if not filename in self._included:
                    self._included.append(filename)

                    self._idl.out('#include "I@filename@"\n',
                                  filename=filename)

        self._idl.out(TMPL_idl_file_footer)

        self._idl = None


    def visitModule(self, node):
        for n in node.definitions():
            n.accept(self)


    def visitForward(self, node):
        xpcom_name = convert_scoped2flat(node.scopedName())

        self._idl.out('interface @xpcomname@;',
                      xpcomname=xpcom_name)


    def visitStruct(self, node):
        corba_name = idlutil.ccolonName(node.scopedName())
        xpcom_name = convert_scoped2flat(node.scopedName())

        uuid = self.computeHexDigest(node.repoId())
        uuid = uuid[:8] + '-' + uuid[8:12] + '-' + uuid[12:16] + '-' + uuid[16:20] + '-' + uuid[20:]

        self._idl.out(TMPL_idl_struct_header,
                      uuid=uuid, intfname=xpcom_name)
        self._idl.inc_indent()

        for member in node.members():
            identifier = member.declarators()[0].identifier()

            type_obj = xpcorba_type.get_type(member.memberType())
            xpidl_type = type_obj.type_xpidl()

            if xpidl_type:
                self._idl.out('readonly attribute @type@ @identifier@;',
                              type=xpidl_type,
                              identifier=identifier)
            else:
                self._idl.out('void get@capidentifier@(@paramspec@);',
                              paramspec=type_obj.decl_xpidl('retval', 0, 1),
                              capidentifier=string.upper(identifier[0]) + identifier[1:])


        self._idl.dec_indent()
        self._idl.out(TMPL_idl_struct_footer)


    def visitInterface(self, node):
        name = node.identifier()
        xpcomintfname = convert_scoped2flat(node.scopedName())

        if node.repoId()[:4] == 'DCE:':
            uuid = string.split(node.repoId()[4:], ':')[0]
        else:
            uuid = self.computeHexDigest(node.repoId())
            uuid = uuid[:8] + '-' + uuid[8:12] + '-' + uuid[12:16] + '-' + uuid[16:20] + '-' + uuid[20:]

        # There is no multiple inheritance in XPIDL. Therefore, in cases where
        # the IDL uses it, always only use the first interface...
        supername = ''
        for i in node.inherits():
            if i.repoId() == REPOID_ISUPPORTS:
                break
            supername = supername + convert_scoped2flat(i.scopedName())
            break
        if supername == '':
            supername = 'nsISupports'
        
        self._idl.out(TMPL_idl_intf_header,
                      uuid=uuid, intfname=xpcomintfname,
                      supername=supername)

        self._idl.inc_indent()

        for c in node.contents():
            c.accept(self)

        self._idl.dec_indent()
        self._idl.out(TMPL_idl_intf_footer)

    def visitTypedef(self, node):
        if not isinstance(node.aliasType().unalias(), idltype.Sequence):
            xpcom_type = xpcorba_type.get_type(node.aliasType())

            for d in node.declarators():
                # Treated specially by XPIDL due to nsrootidl.idl
                if (d.identifier() == 'DOMString'):
                    continue
                # If the alias type has the same name as the aliases, we just
                # ignore the typedef. This allows typedefs designed to forward
                # definitions between modules to work in XPIDL (which doesn't
                # have the concept of modules).
                if isinstance(node.aliasType(), idltype.Declared):
                    if d.identifier() == node.aliasType().scopedName()[-1]:
                        continue
                self._idl.out(TMPL_idl_typedef,
                              alias_type=xpcom_type.type_xpidl(),
                              name=d.identifier())


    def visitConst(self, node):
        scoped_name = node.scopedName()[:-1]

        if len(scoped_name) and isinstance(idlast.findDecl(scoped_name),
                                           idlast.Interface):
            name = node.identifier()

            xpcom_type = xpcorba_type.get_type(node.constType())
 
            cleaned_value = str(node.value())
            if cleaned_value[-1] == 'L':
                cleaned_value = cleaned_value[0:-1]

            self._idl.out(TMPL_idl_const,
                          decl=xpcom_type.decl_xpidl(name, 0, 0),
                          value=cleaned_value)


    def visitEnum(self, node):
        scoped_name = node.scopedName()[:-1]

        if len(scoped_name) and isinstance(idlast.findDecl(scoped_name),
                                           idlast.Interface):
            value = 0

            for enum in node.enumerators():
                name = enum.identifier()

                self._idl.out(TMPL_idl_const,
                              decl='unsigned long ' + name,
                              value=str(value))

                value = value + 1


    def visitAttribute(self, node):
        xpcom_type = xpcorba_type.get_type(node.attrType())

        for identifier in node.identifiers():
            if node.readonly():
                readonly = 'readonly '
            else:
                readonly = ''

            self._idl.out(TMPL_idl_attribute,
                          readonly=readonly,
                          decl=xpcom_type.decl_xpidl(identifier, 0, 0))


    def visitOperation(self, node):
        name = node.identifier()


        paramspec = []
        for p in node.parameters():
            xpcom_type = xpcorba_type.get_type(p.paramType())

            paramspec.append(xpcom_type.decl_xpidl(p.identifier(),
                                                   p.is_in(), p.is_out()))

        rettype = node.returnType()
        if not (isinstance(rettype, idltype.Base) and
                (rettype.kind() == idltype.tk_void)):
            xpcom_type = xpcorba_type.get_type(node.returnType())
            retval = xpcom_type.decl_xpidl('retval', 0, 1)

            if (string.find(retval, ',') == -1) and (string.find(retval, '[') == -1):
                retval = retval[4:-7]
            else:
                paramspec.append(retval)
                retval = 'void'
        else:
            retval = 'void'

        self._idl.out(TMPL_idl_operation,
                      name=name, retval=retval,
                      paramspec=string.join(paramspec, ', '))


def run(tree, args):
    visitor = XPIDL_Visitor()
    tree.accept(visitor)
